def summarize_ppc(target, data, trace_df, m,
                  years=None, cwd=None, dst=None, sigma_colname='sigma', replace_files=False, include_observed=True,
                  quantiles=[0.0, 0.005, 0.025, 0.5, 0.975, 0.995, 1.0],
                  other_columns=None):
    """Analyze an MCMC trace: derive the untreated posterior predictive trajectory.
    
    This function summarizes a trace dataframe into
    a description of the mean and various probability intervals
    of the posterior predictive trajectory of the target society.

    Positional arguments:
        target - (str) title of target society in df's columns
        data - (pandas dataframe) dataset originally analyzed
        trace_df - (pandas dataframe) sampling output for the appropriate variables
        m - (int) number of factors

    Optional keyword arguments:
        years - (numpy array) subset of data's index; full index by default
        cwd - (str) current working directory path; inferred by default
        dst - (str) filepath where to save function outputs; when not specified, nothing saved on disk
        sigma_colname - (str) name of the noise standard deviation varaible in trace_df
        replace_files - (bool) overwrite pre-existing output files; False by default
        include_observed - (bool) include observed data in the output df; True by default
        quantiles - (list/numpy array) quantiles to capture; range, 99%-CI, 95%-CI, and median by default
        other_columns - (iterable of (str, array)) additional columns to include

    Output: pandas dataframe
    
    """
    # Protect trace object from side effects
    trace_df = trace_df.copy()

    # Parameterize country selection & time period
    all_countries = data.columns.to_list()
    years = years if years else data.index.values
    start, end = years[0], years[-1] + 1
    n = len(years)
    order = all_countries.index(target)

    # Other parameterization
    cwd = cwd if cwd else Path(__file__).parent.resolve()
    varnames = trace_df.columns.to_list()
    sigma_position = varnames.index(sigma_colname)
    varnames_short = varnames[:sigma_position] + varnames[sigma_position+1:]
    replacements = ((0.0, 'min'), (1.0, 'max'), (0.5, 'median'))

    # Convert from posterior to predictive posterior

    # Generate noise
    chainlen = trace_df.shape[0]
    sigmas = trace_df[sigma_colname].values
    generate_noise = lambda i: np.random.normal(loc=0.0, scale=sigmas[i], size=n)
    noise_vectors = [generate_noise(i) for i in range(chainlen)]
    noise = np.array(noise_vectors)

    # Combine noise and posterior values
    df = trace_df[varnames_short]
    df = df + noise
    df.columns = years

    # Record summary statistics

    # Calculate summary stats
    mean = df.mean()
    ci_list = [df.quantile(quantile) for quantile in quantiles]

    # Capture summary statistics into a new dataframe
    statistics = [mean] + ci_list
    summary = pd.concat(statistics, axis=1)

    # Generate names for the summary statistic variables
    quantile_names = ['ci_' + str(ci) for ci in quantiles]
    stat_names = ['mean'] + quantile_names
    for number, word in replacements:
        number_name = 'ci_' + str(number)
        if number_name in stat_names:
            index = stat_names.index(number_name)
            stat_names[index] = word
    summary.columns = stat_names

    # Include observed data
    if include_observed:
        summary['observed'] = data[target]

    # Include other specified columns
    if other_columns:
        try:
            for name, vector in other_columns:
                summary[name] = vector
        except TypeError:
            msg = "'other_columns' must contain (str, array) pairs"
            raise TypeError(msg)

    # Optionally, save summary on disk

    if dst:
        if dst == 'default':
            dst = os.path.join(cwd, 'BSCAnalysis', 'L', str(m) + target, 'ppc_summary.pkl')
        dstdir = os.path.split(dst)[0]
        if not os.path.exists(dstdir):
            os.makedirs(dstdir)
        elif os.path.exists(dst) and not replace_files:
            msg = "Summary file already exists. Switch 'replace_files=True'."
            raise Exception(msg)
        with open(dst, 'wb') as file:
            pickle.dump(summary, file)

    # Return a dataframe

    return summary




def plot_factors(dfs, data, target,
                 dstdir=None, filename=None, alpha=0.333, ci_lo='ci_0.025', ci_hi='ci_0.975', yscale=0,
                 line_avg={'linestyle': '-', 'color': 'k', 'label': 'Mean predicted', 'linewidth': 1.0},
                 xlabel='Year', ylabel='Value', title=None, image_format='pdf', output='disk',
                 verbose=True):
    """Visualize the latent factor trajectories.
    
    This function turns a list of pandas datframes of summary statistics
    of the latent factors' posterior distributions
    into a matplotlib visualization.

    Positional arguments:
        dfs - (list) list of df-s of summary statistics for factors
        data - (pandas dataframe) dataset originally analyzed
        target - (str) title of target society in df's columns

    Optional keyword arguments:
        dstdir - (str) destination path for plot output directory; automatically constructed by default
        filename - (str) destination path for plot output filename; automatically constructed by default 
        alpha - (float) shading transparency of the confidence interval; 0.333 by default
        ci_lo - (str) variable/column name of the lower bound of the confidence interval; summarize_ppc()'s default by default
        ci_hi - (str) variable/column name of the upper bound of the confidence interval; summarize_ppc()'s default by default
        yscale - (int) orders of magnitude for scaling up the y-variable upon plotting; 0 by default
        line_avg - (dict) matplotlib kwargs for formatting the posterior mean trajectory; preconfigured by default
        xlabel - (str) label of X-axis; 'Year' by default
        ylabel - (str) label of Y-axis; 'Value' by default
        title - (str) title of the figure; None by default
        image_format - (str) file format of figure file; 'pdf' by default
        output - (str) 'disk' (save output on disk; default) or 'display' (display figure to user)
        verbose - (bool) print messages to user; True by default

    Output: None
    
    """
    #Configure time period
    years = data.index.values
    zeros = '' if (yscale < 1) else " '" + ''.join(['0' for i in range(yscale)])
    coefficient = 10**yscale
    m = len(dfs)

    # Create and index matplotlib subfigures
    row_num, col_num = int(np.ceil(m/2)), 2
    rows, cols = range(row_num), range(col_num)
    figure, axes = plt.subplots(row_num, col_num, sharex='col')
    indexes = list(itertools.product(rows, cols))

    # Iterate over factors
    for i, index in enumerate(indexes):
        # Capture appropriate factor data and matplotlib subfigure
        df, axis = dfs[i].copy(), axes[index]
        # Adjust the y-scale of the graphs
        df /= coefficient
        # Visualize the posterior trajectory
        low, high, mean = df[ci_lo], df[ci_hi], df['mean']
        axis.fill_between(years, low, high, color='tab:blue', alpha=alpha)
        axis.plot(years, mean, **line_avg)
        axis.set_title('Factor #' + str(i))
        axis.set_ylabel(ylabel + zeros)

    # Top-level visual layout
    figure.tight_layout()
    # If configured so, save the graph on disk
    if output == 'disk':
        # Configure file paths
        filename = filename if filename else 'factors.' + image_format
        if not dstdir:
            cwd = Path(__file__).parent.resolve()
            dstdir = os.path.join(cwd, 'BSCFigures', 'L' + str(m) + target)
        # Create directory path if necessary
        if not os.path.exists(dstdir):
            os.makedirs(dstdir)
        # Save the figure
        dst = os.path.join(dstdir, filename)
        plt.savefig(dst, bbox_inches='tight')
        plt.close()
        # Report to user
        if verbose:
            msg = 'Figure saved: {}'.format(dst)
            print(msg)
    # Otherwise simply display the figure to user
    elif output == 'display':
        plt.show()
    else:
        msg = "Option 'output={}' not supported".format(output)
        raise NotImplementedError(msg)

    return None




import os
import shutil
import pickle
import itertools
import numpy as np
import pandas as pd
import seaborn as sns
from pathlib import Path
import matplotlib.pyplot as plt

from .plotppc import plot_ppc
from .plotppcdiff import plot_ppc_diff
from .plotppcbell import plot_ppc_bell
from .plotfactors import plot_factors



import os
import shutil
import pickle
import itertools
import numpy as np
import pandas as pd
from pathlib import Path

from .readtrace import read_tracefile
from .summarizeppc import summarize_ppc
from .summarizefactors import summarize_factors



def plot_ppc_bell(tracedf, data, target, m,
                  difference=True, samplesize='full', year_pos=None, sigma_varname='sigma', dstdir=None, filename=None, alpha=0.333,
                  line_kde={'color': 'k', 'label': '', 'linewidth': 1.3},
                  shade_kde={'color': 'tab:blue', 'label': '', 'shade': True},
                  line_obs={'linestyle': '--', 'color': 'k'},
                  bandwidth='scott', xlabel='Value', ylabel='Probability density', title=None, image_format='pdf', output='disk',
                  verbose=True):
    """Visualize the target society's cumulative treatment effect.
    
    This function turns a trace dataframe
    (with appropriate variables present)
    into a visualization of the posterior predictive distribution
    of the size of the cumulative treatment effect for the target society
    at a specified year (end of time period by default).

    Positional arguments:
        tracedf - (pandas dataframe) sampling output for the appropriate variables
        data - (pandas dataframe) dataset originally analyzed
        target - (str) title of target society in df's columns
        m - (int) number of factors; 4 by default

    Optional keyword arguments:
        difference - (bool) visualize treatment effect instead of absolute ppc value; True by default
        samplsize - (str or int) 'full' to use the full trace (default); integer to select ransom subsample
        year_pos - (int) position of the target year within  data's index (-1 ie. last year by default)
        sigma_varname - (str) name of the noise standard deviation varaible in tracedf; 'sigma' by default
        dstdir - (str) destination path for plot output directory; automatically constructed by default
        filename - (str) destination path for plot output filename; automatically constructed by default 
        alpha - (float) shading transparency of the confidence interval; 0.333 by default
        line_kde - (dict) matplotlib kwargs for formatting density plot's outline; preconfigured by default
        shade_kde - (dict) matplotlib kwargs for formatting density plot's shade; preconfigured by default
        line_obs - (dict) matplotlib kwargs for formatting observed value marker; preconfigured by default
        bandwidth - (str or float) density plot's bandwidth parameter or seaborn selection method therefor; 'scott' by default
        xlabel - (str) label of X-axis; 'Year' by default
        ylabel - (str) label of Y-axis; 'Probability density'
        title - (str) title of the figure; None by default
        image_format - (str) file format of figure file; 'pdf' by default
        output - (str) 'disk' (save output on disk; default) or 'display' (display figure to user)
        verbose - (bool) print messages to user; True by default

    Output: None
    
    """
    # Protect source data from side effects
    observed = data.copy()

    # Configure choice of year
    year_pos = year_pos if year_pos else -1
    year = observed.index[year_pos]
    # Record observed value
    observation = observed[target].iloc[year_pos]
    # Carve out relevant sections of the trace
    sigma_trace = tracedf[sigma_varname]
    tracedf.drop(sigma_varname, axis=1, inplace=True)
    tracedf = tracedf.iloc[:, year_pos]
    # Carve out a small sample of the trace
    chainlen = tracedf.shape[0]
    samplesize = chainlen if (samplesize == 'full') else samplesize
    selection = np.random.choice(np.arange(chainlen), size=samplesize, replace=False)
    tracedf, sigma = tracedf.iloc[selection], sigma_trace[selection]
    # Convert from posterior parameter trace to posterior predictive trace
    noise = np.random.normal(0, sigma)
    ppc = tracedf + noise
    # Convert to treatment effects if appropriate
    if difference:
        ppc = observation - ppc
    # Draw up the baseline graph
    sns.kdeplot(ppc, bw=bandwidth, **line_kde)
    sns.kdeplot(ppc, bw=bandwidth, alpha=alpha, **shade_kde)
    # Draw observed value/0 treatment effect vertical line
    null_value = 0 if difference else observation
    plt.axvline(null_value, **line_obs)
    # Add labels, and optionally add a title
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    if title:
        plt.title(title)
    # If configured so, save the graph on disk
    if output == 'disk':
        # Configure file paths
        filename = filename if filename else 'ppc_bell_' + str(year) + '.' + image_format
        if not dstdir:
            cwd = Path(__file__).parent.resolve()
            dstdir = os.path.join(cwd, 'BSCFigures', 'L' + str(m) + target)
        # Create directory path if necessary
        if not os.path.exists(dstdir):
            os.makedirs(dstdir)
        # Save the figure
        dst = os.path.join(dstdir, filename)
        plt.savefig(dst, bbox_inches='tight')
        plt.close()
        # Report to user
        if verbose:
            msg = 'Figure saved: {}'.format(dst)
            print(msg)
    # Otherwise simply display the figure to user
    elif output == 'display':
        plt.show()
    else:
        msg = "Option 'output={}' not supported".format(output)
        raise NotImplementedError(msg)
    return None




def plot_ppc(summary, cutoff, target,
             dstdir=None, filename=None, m=None, alpha=0.333, ci_lo='ci_0.025', ci_hi='ci_0.975', ylim=None,
             line_avg={'linestyle': '--', 'color': 'k', 'label': 'Mean predicted'},
             line_obs={'linestyle': '-', 'color': 'k', 'label': 'Observed'},
             line_cutoff={'linestyle': ':', 'color': 'k'},
             xlabel='Year', ylabel='Value', title=None, image_format='pdf', output='disk',
             verbose=True):
    """Visualize the target society's posterior predictive trajectory.
    
    This function turns a pandas datframe of summary statistics
    for the target society's posterior predictive trajectory
    into a matplotlib visualization.

    Positional arguments:
        summary - (pandas dataframe) df of summary statistics for the target trajectory
        cutoff - (int) the start year of treatment for target society
        target - (str) title of target society in df's columns

    Optional keyword arguments:
        dstdir - (str) destination path for plot output directory; automatically constructed by default
        filename - (str) destination path for plot output filename; automatically constructed by default 
        m - (int) number of factors; 4 by default
        alpha - (float) shading transparency of the confidence interval; 0.333 by default
        ci_lo - (str) variable/column name of the lower bound of the confidence interval; summarize_ppc()'s default by default
        ci_hi - (str) variable/column name of the upper bound of the confidence interval; summarize_ppc()'s default by default
        line_avg - (dict) matplotlib kwargs for formatting the posterior mean trajectory; preconfigured by default
        line_obs - (dict) matplotlib kwargs for formatting the observed trajectory; preconfigured by default
        line_cutoff - (dict) matplotlib kwargs for formatting the cutoff year marker; preconfigured by default
        xlabel - (str) label of X-axis; 'Year' by default
        ylabel - (str) label of Y-axis; 'Value' by default
        title - (str) title of the figure; None by default
        image_format - (str) file format of figure file; 'pdf' by default
        output - (str) 'disk' (save output on disk; default) or 'display' (display figure to user)
        verbose - (bool) print messages to user; True by default

    Output: None
    
    """
    # Protect summary object from side effects
    summary = summary.copy()
    # Parameterize
    years = summary.index
    # Draw up the baseline graph
    low, high, mean = summary[ci_lo], summary[ci_hi], summary['mean']
    plt.fill_between(years, low, high, alpha=alpha)
    plt.plot(years, mean, **line_avg)
    # If observed data is available, plot it
    try:
        observed = summary['observed']
        plt.plot(years, observed, **line_obs)
    except KeyError:
        pass
    # Add labels and legend, and optionally add a title and adjust the vertical scale
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.legend()
    if ylim:
        plt.ylim(ylim)
    if title:
        plt.title(title)
    # Visualize start year of treatment
    plt.axvline(cutoff, **line_cutoff)
    # If configured so, save the graph on disk
    if output == 'disk':
        # Configure file paths
        filename = filename if filename else 'ppc_trajectory.' + image_format
        if not dstdir:
            cwd = Path(__file__).parent.resolve()
            dstdir = os.path.join(cwd, 'BSCFigures', 'L' + str(m) + target)
        # Create directory path if necessary
        if not os.path.exists(dstdir):
            os.makedirs(dstdir)
        # Save the figure
        dst = os.path.join(dstdir, filename)
        plt.savefig(dst, bbox_inches='tight')
        plt.close()
        # Report to user
        if verbose:
            msg = 'Figure saved: {}'.format(dst)
            print(msg)
    # Otherwise simply display the figure to user
    elif output == 'display':
        plt.show()
    else:
        msg = "Option 'output={}' not supported".format(output)
        raise NotImplementedError(msg)
    return None




import os
import shutil
import pickle
import numpy as np
import pymc3 as pm
import pandas as pd
from pathlib import Path
import theano.tensor as T
from sklearn.decomposition import PCA

from .fit import fit



def summarize_factors(target, data, trace_df, m,
                      years=None, cwd=None, dst=None, replace_files=False,
                      quantiles=[0.0, 0.005, 0.025, 0.5, 0.975, 0.995, 1.0]):
    """Analyze an MCMC trace: derive the latent factor distributions.
    
    This function summarizes a trace dataframe into
    a description of the mean and various probability intervals
    of the latent factor trajectories of.

    Positional arguments:
        target - (str) title of target society in df's columns
        data - (pandas dataframe) dataset originally analyzed
        trace_df - (pandas dataframe) sampling output for the appropriate variables
        m - (int) number of factors

    Optional keyword arguments:
        years - (numpy array) subset of data's index; full index by default
        cwd - (str) current working directory path; inferred by default
        dst - (str) filepath where to save function outputs; when not specified, nothing saved on disk
        replace_files - (bool) overwrite pre-existing output files; False by default
        quantiles - (list/numpy array) quantiles to capture; range, 99%-CI, 95%-CI, and median by default

    Output: pandas dataframe
    
    """
    # Protect trace object from side effects
    trace_df = trace_df.copy()

    # Parameterize country selection & time period
    years = years if years else data.index.values
    start, end = years[0], years[-1] + 1
    n = len(years)

    # Other parameterization
    cwd = cwd if cwd else Path(__file__).parent.resolve()
    replacements = ((0.0, 'min'), (1.0, 'max'), (0.5, 'median'))

    # Initialize storage
    dflist = []
    
    # Iterate over factors
    for i in range(m):
        # Carve out data for the i-th factor
        mslice = slice(i, m*n, m)
        df_m = trace_df.iloc[:, mslice]
        # Calculate summary statistics
        mean = df_m.mean()
        ci_list = [df_m.quantile(quantile) for quantile in quantiles]
        # Combine statistics in a single dataframe
        statistics = [mean] + ci_list
        summary = pd.concat(statistics, axis=1)
        # Give each statistic a title
        quantile_names = ['ci_' + str(ci) for ci in quantiles]
        stat_names = ['mean'] + quantile_names
        # Give intuitive names to a few special statistics (min, max, median)
        for number, word in replacements:
            number_name = 'ci_' + str(number)
            if number_name in stat_names:
                index = stat_names.index(number_name)
                stat_names[index] = word
        # Name columns and rows of the summary df, and append it to storage
        summary.columns = stat_names
        summary.index = years
        dflist.append(summary)

    # Optionally, save summary on disk

    if dst:
        if dst == 'default':
            dst = os.path.join(cwd, 'BSCAnalysis', 'L', str(m) + target, 'factor_summary.pkl')
        dstdir = os.path.split(dst)[0]
        if not os.path.exists(dstdir):
            os.makedirs(dstdir)
        elif os.path.exists(dst) and not replace_files:
            msg = "Summary file already exists. Switch 'replace_files=True'."
            raise Exception(msg)
        with open(dst, 'wb') as file:
            pickle.dump(dflist, file)

    # Return the list of dataframes

    return dflist




from setuptools import setup, find_packages

with open('README.md', 'r') as file:
    long_description = file.read()

setup(name='bayessynth',
      version='0.1',
      url='https://github.com/eliastuo/bayessynth',
      license='MIT',
      author='Elias Tuomaala',
      author_email='mail@eliastuomaala.com',
      description='Python implementation of Bayesian Synthetic Controls',
      packages=find_packages(),
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
      python_requires='>=3.7.4')



def fit(df, target, cutoff, prior,
        years=None, countries=None, llambda_raw=2, m=4,
        cwd=None, out_dst=None,
        chains=2, chain_length=25000, tune_length=5000, target_accept=0.90, treedepth=12, samplesize=2000, progressbar=True, mcmc_kwargs={},
        replace_files=False, verbose=True, trace_only=False):
    """Main function to estimate the BSC model for a single dataset.

    The BSC model is fitted on the dataset provided,
    using th pymc3 implementation of the NUTS (MCMC) sampler.
    The resulting sample trace is stored on hard drive
    along with various other summary statistics.
    
    Positional arguments:
        df - (pandas dataframe) data to be analyzed
        target - (str) title of target society in df's columns
        cutoff - (int) the start year of treatment for target society
        prior - (dict) prior distribution parameters

    Optional keyword arguments:
        years - (numpy array) subset of df's index; full index by default
        countries - (list) subset of countries in df's columns; all columns by default
        llambda_raw - (int) scaling parameter for factor prior standard devations; 2 by default
        m - (int) number of factors; 4 by default
        cwd - (str) current working directory path; inferred by default
        out_dst - (str) destination path for sampler output; automatically constructed by default
        chains - (int) number of parallel MCMC chains; 2 by default
        chain_length - (int) number of proper sampler steps per chain; 25000 by default
        tune_length - (int) number of additional tune-in steps per chain; 5000 by default
        target_accept - (float) NUTS sampler target acceptance rate; 0.9 by default
        treedepth - (int) NUTS sampler maximum tree depth; 12 by default
        samplesize - (int) size of a separately saved representative mini-sample; 2000 by default
        progressbar - (bool) print a live progressbar while sampling; True by default
        mcmc_kwargs - (dict) additional keyword arguments for the NUTS sampler
        replace_files - (bool) overwrite pre-existing sampler output files; False by default
        verbose - (bool) print messages to user; True by default
        trace_only - (bool) save only sampling trace, no summaries; False by default

    Output: None
    """

    # Parameter setup

    # Set up time period
    years = years if years else df.index.values
    start, end = years[0], years[-1] + 1
    years_short = np.arange(start, cutoff)
    years_rest = np.arange(cutoff, end)
    n, n_short, n_rest = len(years), len(years_short), len(years_rest)
    # Set up countries
    countries = countries if countries else df.columns.tolist()
    countries = [country for country in countries if country != target]
    all_countries = countries + [target]
    d = len(all_countries)
    # File paths
    cwd = cwd if cwd else Path(__file__).parent.resolve()
    if not out_dst:
        out_dst = os.path.join(cwd, 'BSCOutput', 'L' + str(m) + target)
    csv = os.path.join(out_dst, 'trace')
    dst_results = os.path.join(out_dst, 'results.pkl')
    dst_modeldict = os.path.join(out_dst, 'modeldict.pkl')
    dst_smallsample = os.path.join(out_dst, 'smallsample.pkl')
    summary = {}
    # Create or replace directories
    for path in [out_dst]:
        if os.path.exists(path):
            if replace_files:
                shutil.rmtree(path)
            else:
                msg = 'Directory "{}" already exists.'.format(path)
                raise Exception(msg)
        os.makedirs(path)
    # Latent factor prior variance
    llambda = llambda_raw**2
    
    # Data structure setup

    # PCA components
    dh = df.copy()[all_countries]
    pca = PCA(n_components=m)
    pca.fit(dh[countries])
    database = pca.transform(dh[countries])
    database_flat = database.reshape(n*m)
    base_var = database.var(axis=0)
    factor_var = np.concatenate([base_var for i in range(n)])
    # Outcome data
    datarest = dh.loc[years, all_countries].values
    # Indicator matrix
    indicator = np.zeros((n, d))
    indicator[n_short:, -1] = 1
    D = indicator.T
    # Pre-reform dummy vector of target society treatment effects
    dummy = np.zeros(n_short)
    # Identity matrices
    identity_md = np.identity(m*d)
    identity_nm = np.identity(n*m)

    # Sampler parameterization
    mcmc_kwargs.update({'chains': chains,
                        'draws': chain_length,
                        'tune': tune_length,
                        'progressbar': progressbar,
                        'target_accept': target_accept,
                        'max_treedepth': treedepth})

    # Run MCMC sampling

    # Open a pymc3 model in context manager
    with pm.Model() as model:
        # Probabilistic model

        # Society fixed effect (K) hierarchical prior
        k_mu = pm.Normal('k_mu', mu=prior['k_mu'], sd=prior['k_sd'])
        k_sd = pm.HalfCauchy('k_sd', beta=prior['k_gamma'])
        k_offset = pm.Normal('k_offset', mu=0, sd=1, shape=d)
        kappa = (k_offset * k_sd) + k_mu
        # Factor loadings' (B) hierarchical prior
        b_mu = pm.Normal('b_mu', mu=prior['b_mu'], sd=prior['b_sd'], shape=m)
        b_sd = pm.HalfCauchy('b_sd', beta=prior['b_gamma'], shape=m)
        b_long = pm.MvNormal('b_long', mu=0, cov=identity_md, shape=m*d)
        b_offset = b_long.reshape((m, d)).T
        beta = ((b_offset * b_sd) + b_mu).T
        # Noise standard deviation prior
        sigma = pm.HalfCauchy('sigma', beta=prior['sigma_gamma'])
        # Latent factor prior
        L_long = pm.MvNormal('L_long', mu=database_flat,
                             cov=identity_nm*factor_var*llambda, shape=n*m)
        L = L_long.reshape((n, m))
        # Year fixed effect prior
        delta_raw = pm.Normal('delta_raw', mu=prior['delta_mu'], sd=prior['delta_sd'], shape=(n, 1))
        delta = delta_raw - delta_raw.mean()
        # Treatment effect prior
        alpha = pm.Normal('alpha', mu=prior['alpha_mu'], sd=prior['alpha_sd'], shape=n_rest)
        alpha_vector = T.concatenate([dummy, alpha])
        alpha_matrix = (D * alpha_vector).T
        # Untreated state
        synthetic = pm.Deterministic('synthetic', T.dot(L, beta) + kappa + delta)
        # Likelihood
        y = pm.Normal('y', mu=synthetic+alpha_matrix, sd=sigma, observed=datarest)

        # MCMC sampling

        # Initialize hard drive storage for sampler
        db = pm.backends.Text(csv)
        # Draw a sample
        trace = pm.sample(**mcmc_kwargs, trace=db)
    
    if not trace_only:

        # Format and save output
        
        # Save model and trace objects
        with open(dst_modeldict, 'wb') as file:
            pickle.dump({'model': model, 'trace': trace}, file)
        # Save a small random sample separately
        samples = np.random.choice(trace, size=samplesize)
        with open(dst_smallsample, 'wb') as file:
            pickle.dump(samples, file)

        # Calculate summary statistics

        # Record observed data and initialize storage variables
        observed = datarest[:, -1]
        greater = np.zeros(n)
        difference = np.zeros(n)
        # Loop over the small sample
        for draw in samples:
            # Record untreated state and resample noise
            mu = draw['synthetic'][:, -1]
            sigma = draw['sigma']
            noise = np.random.normal(loc=0.0, scale=sigma, size=n)
            # Construct draw from posterior predictive and compare to observed data
            prediction = mu + noise
            greater += (observed >= prediction)
            difference += (observed - prediction)
        # Calculate the average difference to observed data
        difference /= samplesize
        # Calculate the share of predictions less than observed data
        greater /= samplesize
        # Calculate (Gelman-Rubin) diagnostic of sampler convergence
        rubin = pm.diagnostics.gelman_rubin(trace, varnames=['alpha'])['alpha']
        # Save summary statistics
        results = [greater, difference, rubin]
        with open(dst_results, 'wb') as file:
            pickle.dump(results, file)

    # Report back to user
    if verbose:
        print("Finished fitting the model.")
        print("The following output files were generated:")
        msg = "\t{}"
        for path in [dst_modeldict, csv, dst_smallsample, dst_results]:
            print(msg.format(path))

    return None




def read_tracefile(target, data, m,
                   trace_dir=None,
                   varnames='ppc', outvarnames=None, years=None, portionsize=1000, cwd=None, tune_length=5000, print_frequency=5, verbose=True):
    """Convert set of hard drive MCMC tracefiles into a single pandas dataframe.
    
    This function reads bit-by-bit into memory the large tracefiles
    generated by bs.fit(), capturing only desired variables.
    Typically, the tracefiles are too large to fit in memory in full,
    so desired variables should be a small subset of the full variable list.

    Positional arguments:
        target - (str) title of target society in df's columns
        data - (pandas dataframe) dataset originally analyzed
        m - (int) number of factors

    Optional keyword arguments:
        trace_dir - (str) source directory of the tracefiles; bs.fit()'s default used by default
        varnames - (str or list) variable selection method ('ppc' or 'factors') or list of varnames
        outvarnames - (list) variable names to overwrite the original varnames
        years - (numpy array) subset of data's index; full index by default
        portionsize - (int) size of the bit of trace read into memory at once; 1000 by default
        cwd - (str) current working directory path; inferred by default
        tune_length - (int) number of additional tune-in steps per chain; bs.fit()'s default by default
        print_frequency - (int) number of portions read without printing updates to user; 5 by default
        verbose - (bool) print messages to user; True by default

    Output: pandas dataframe
    
    """

    # Parameterize country selection & time period
    all_countries = data.columns.to_list()
    years = years if years else data.index.values
    n = len(years)
    order = all_countries.index(target)

    # Parameterize file paths and variable selection
    cwd = cwd if cwd else Path(__file__).parent.resolve()
    if not trace_dir:
            trace_dir = os.path.join(cwd, 'BSCOutput', 'L' + str(m) + target, 'trace')
    tracefiles = os.listdir(trace_dir)
    chains = len(tracefiles)
    tracefiles = [os.path.join(trace_dir, tracefile) for tracefile in tracefiles]

    # Select appropriate vairables
    if type(varnames) == list:
        if not all(type(statname) == str for statname in statnames):
            msg = "'varnames' must be str or list of str's"
            raise AttributeError(msg)
    elif varnames == 'ppc':
        varnames = [('synthetic__' + str(i) + '_' + str(order)) for i in range(n)] + ['sigma']
    elif varnames == 'factors':
        varnames = [('L_long__' + str(number)) for number in range(n*m)]
    elif type(varnames) != str:
        msg = "'varnames' must be list or str"
        raise AttributeError(msg)
    else:
        msg = "Variable selection method '{}' not recognised".format(varnames)
        raise NotImplementedError(msg)

    # Initialize pandas access to the trace files on hard drive
    readers = []
    for tracefile in tracefiles:
        reader = pd.read_csv(tracefile, chunksize=portionsize)
        readers.append(reader)

    # Initialize storage for carving out the relevant data sections
    dflists = [[] for tracefile in tracefiles]

    # Print target country to user
    if verbose:
        print('Reading tracefile of {}'.format(target))

    # Read data from hard drive in chunks and carve out relevant sections
    for i, portions in enumerate(zip(*readers)):
        # Skip the tune-in section of the trace
        if (i + 1) * portionsize <= tune_length:
            continue
        # Occasionally report progress to user
        if verbose and (i % print_frequency == 0):
            msg = '\tReached chunk #{}: draws {}-{}'.format(i, i*portionsize, (i+1)*portionsize)
            print(msg)
        # Record the relevant portion of each trace in memory
        for dflist, portion in zip(dflists, portions):
            dflist.append(portion[varnames])

    # Concatenate portions by trace
    data_by_trace = []
    for dflist in dflists:
        data_by_trace.append(pd.concat(dflist))
    # Aggregate data from all traces
    df = pd.concat(data_by_trace, ignore_index=True)
    if outvarnames:
        df.columns = outvarnames

    return df

